<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2019 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2019.1/1829459
 */

namespace Reviews\Validator;

use Application\Config\ConfigManager;
use Application\Option;
use Comments\Model\Comment;
use Reviews\Model\Review;
use Projects\Model\Project;
use SebastianBergmann\CodeCoverage\InvalidArgumentException;
use Users\Validator\Users;

class Transitions
{
    const OPT_UP_VOTES = 'includeExtraUpVotes';
    private $review;
    private $options;
    private $p4Admin;
    private $transitions;

    /**
     * Transitions constructor.
     * @param Review    $review     the current review
     * @param array     $options    Supported values:
     *                                  OPT_UP_VOTES     - Implicit up votes to include
     *                                  Option::USER_ID  - the current authenticated user ID
     *                                  Option::IS_SUPER - whether the user is super
     *                                  Option::SERVICES - The services provider
     */
    public function __construct(Review $review, $options = array())
    {
        $defaults = array(
            self::OPT_UP_VOTES => null,
            Option::IS_SUPER   => false,
            Option::USER_ID    => null,
            Option::SERVICES   => null
        );

        Option::validate($options, $defaults);
        $this->review  = $review;
        $this->options = $options;
        $services      = $options[Option::SERVICES];
        if (is_null($services)) {
            throw new \InvalidArgumentException(
                "Services is required!"
            );
        }
        $this->p4Admin     = $services->get('p4_admin');
        $this->transitions = $review->getTransitions();
    }

    /**
     * Merge the up-voters with the author
     * @return array Return a list of valid user account.
     * @throws \P4\Spec\Exception\NotFoundException
     */
    private function populateUpVotes()
    {
        $includeExtraUpVotes =
            array($this->review->isValidAuthor()
                ? $this->review->getAuthorObject()->getId()
                : $this->review->get('author'));

        $upVoters = $this->options[self::OPT_UP_VOTES];
        // Take a list of user name or group names and put them into a array
        if ($upVoters && !empty($upVoters)) {
            $userValidator = new Users(array('connection' => $this->p4Admin));
            if ($userValidator->isValid($upVoters)) {
                $includeExtraUpVotes = array_merge($includeExtraUpVotes, $upVoters);
            } else {
                $messages = $userValidator->getMessages();
                throw new \InvalidArgumentException($messages[Users::UNKNOWN_IDS]);
            }
        }
        return $includeExtraUpVotes;
    }

    /**
     * Get transitions for review model and filter response.
     * If the user isn't a candidate to transition this review, false is returned. It is recommended
     * the transition UI be disabled in that case.
     * If the user is a candidate, an array will be returned though it may be empty. Even an empty
     * array indicates transitioning is viable and the UI may opt to stay enable and show items such
     * as 'add a commit' in this case.
     *
     * @return array array of available transitions (may be empty) or false
     * @throws \Application\Config\ConfigException
     * @throws \P4\Exception
     * @throws \P4\Spec\Exception\NotFoundException
     * @throws \Record\Exception\NotFoundException
     * @throws \Exception
     */
    public function getAllowedTransitions()
    {
        $p4Admin             = $this->p4Admin;
        $userId              = $this->options[Option::USER_ID];
        $includeExtraUpVotes = $this->populateUpVotes();

        $members    = array();
        $moderators = array();
        $impacted   = $this->review->getProjects();
        $this->populateImpactedMembersAndModerators($moderators, $members, $impacted);
        // use the stringMatches helper as it accounts for the server's case sensitivity
        $isAuthor = $p4Admin->stringMatches($userId, $this->review->get('author'));
        $this->checkDisableCommits();
        $this->checkAuthorAllowedToCommit($isAuthor);
        $this->checkOutstandingVotes($includeExtraUpVotes, $impacted);
        $this->checkOutstandingTasks();
        $this->checkRoles(
            $moderators,
            $members,
            $isAuthor,
            $impacted
        );
        return $this->transitions;
    }

    /**
     * Get a list of moderators and members that will impact the transition of this review.
     *
     * @param array       $impacted     This is a list of project this review impacts.
     * @param array       $moderators   This is the list of moderators for the review.
     * @param array       $members      This is the members of the impacted projects
     *
     * @throws \Exception
     */
    private function populateImpactedMembersAndModerators(&$moderators, &$members, $impacted)
    {
        // prepare list of members/moderators of review-related projects/project-branches
        if ($impacted) {
            $projects = Project::fetchAll(array('ids' => array_keys($impacted)), $this->p4Admin);
            foreach ($projects as $project) {
                $branches = $impacted[$project->getId()];
                foreach ($branches as $branch) {
                    foreach ($project->getModerators(array($branch)) as $branchModerator) {
                        $moderators = array_merge_recursive(
                            $moderators,
                            array($project->getId() => array($branch => array($branchModerator)))
                        );
                    }
                }
                $members = array_merge($members, $project->getAllMembers());
            }
        }
    }

    /**
     * Check if the disable commits is enabled.
     * @throws \Application\Config\ConfigException
     */
    private function checkDisableCommits()
    {
        $services = $this->options[Option::SERVICES];
        // remove option to commit a review if disabled in route params or config
        $disableCommitConfig = ConfigManager::getValue(
            $services->get(Option::CONFIG),
            ConfigManager::REVIEWS_DISABLE_COMMIT,
            false
        );
        if ($disableCommitConfig) {
            unset($this->transitions[Review::STATE_APPROVED_COMMIT]);
        }
    }

    /**
     * Check if we allow the Author to commit the review or not.
     *
     * @param boolean       $isAuthor      Check to see if current user is the author of the review.
     *
     * @throws \Application\Config\ConfigException
     */
    private function checkAuthorAllowedToCommit($isAuthor)
    {
        $services           = $this->options[Option::SERVICES];
        $disableSelfApprove = ConfigManager::getValue(
            $services->get(Option::CONFIG),
            ConfigManager::REVIEWS_DISABLE_SELF_APPROVE,
            false
        );
        // deny authors approving their own reviews if self-approval is disabled in config
        // note this will take place even if the author is also a moderator
        if ($isAuthor && $this->review->get('state') !== Review::STATE_APPROVED && $disableSelfApprove == true) {
            unset($this->transitions[Review::STATE_APPROVED]);
            unset($this->transitions[Review::STATE_APPROVED_COMMIT]);
        }
    }

    /**
     * Check if we have any outstanding votes that are required. including self and any extra
     * up voters we might have passed it that will meet the requirement.
     *
     * @param array        $includeExtraUpVotes  A list of user that if had voted could change outcome of transitions
     * @param array        $impacted             A list of project with branches.
     * @throws \Application\Config\ConfigException
     * @throws \P4\Exception
     * @throws \Record\Exception\NotFoundException
     */
    private function checkOutstandingVotes($includeExtraUpVotes, $impacted)
    {
        // Do this after the more simple checks as there is no need to check all the votes
        // if approval is already not permitted.
        // if we have required reviewers, and they haven't all
        // up voted don't allow anyone to approve or commit unless
        // the current user is the last required reviewer to vote, then we
        // will let them approve, and take their approval as an up-vote
        // This call passing in the $userId tests for outstanding votes
        // assuming $userId votes up also
        // It is possible from P4V etc for the author to be added as a required
        // reviewer (we filter these out in the UI). We disregard the author vote
        // by assuming a vote up as we do not allow authors to vote on their
        // own reviews. Without assuming their vote other users will not be
        // able to approve and commit if the authors vote is required
        if ((isset($this->transitions[Review::STATE_APPROVED]) ||
                isset($this->transitions[Review::STATE_APPROVED_COMMIT])) &&
            !$this->review->canApprove(
                $includeExtraUpVotes,
                $impacted,
                $this->options[Option::SERVICES]
            )
        ) {
            unset($this->transitions[Review::STATE_APPROVED]);
            unset($this->transitions[Review::STATE_APPROVED_COMMIT]);
        }
    }

    /**
     * Check to see if we are gating the review based on outstanding comments marked as
     * tasks that need to be addressed.
     */
    private function checkOutstandingTasks()
    {
        try {
            $services               = $this->options[Option::SERVICES];
            $outstandingTasksConfig = ConfigManager::getValue(
                $services->get(Option::CONFIG),
                ConfigManager::REVIEWS_DISABLE_APPROVE_WHEN_TASKS_OPEN,
                false
            );
        } catch (\Exception $e) {
            // Ignore config errors and allow the remaining transition checks to process.
            $outstandingTasksConfig = false;
        }
        // Check to see if open tasks should prevent approval
        if ((isset($this->transitions[Review::STATE_APPROVED]) ||
                isset($this->transitions[Review::STATE_APPROVED_COMMIT]))
            && $outstandingTasksConfig) {
            try {
                $comments = Comment::fetchAll(
                    array(
                        Comment::FETCH_BY_TOPIC        => array('reviews/' . $this->review->getId()),
                        Comment::FETCH_BY_TASK_STATE   => array(Comment::TASK_OPEN),
                        Comment::FETCH_IGNORE_ARCHIVED => true
                    ),
                    $this->p4Admin
                );
            } catch (\Exception $e) {
                // We failed to fetch comments.
                $comments = false;
            }
            if ($comments) {
                $commentsArray = $comments->toArray(true);
                if (!empty($commentsArray)) {
                    unset($this->transitions[Review::STATE_APPROVED]);
                    unset($this->transitions[Review::STATE_APPROVED_COMMIT]);
                }
            }
        }
    }

    /**
     * Tests the populated moderators to see if the current user is mentioned
     * @param array $moderators
     * @return bool true if the current user is found
     */
    private function isModerator($moderators)
    {
        $p4Admin = $this->p4Admin;
        $userId  = $this->options[Option::USER_ID];
        foreach ((array)$moderators as $project => $branches) {
            foreach ($branches as $branch => $mods) {
                if ($p4Admin->stringMatches($userId, $mods)) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Filter the transitions base on the users role as part of the review.
     * @param array     $moderators     This is the list of moderators for the review, and the relevant branches
     * @param array     $members        Project members across the review
     * @param boolean   $isAuthor       Check to see if current user is the author of the review.
     * @param array     $impacted       This is a list of project this review impacts.
     *
     * @throws \Application\Config\ConfigException
     * @throws \Exception
     */
    private function checkRoles(
        $moderators,
        $members,
        $isAuthor,
        $impacted
    ) {
        $isSuper     = $this->options[Option::IS_SUPER];
        $p4Admin     = $this->p4Admin;
        $userId      = $this->options[Option::USER_ID];
        $isMember    = $p4Admin->stringMatches($userId, $members);
        $isModerator = $this->isModerator($moderators);
        $isApprover  = in_array($this->review->getHeadVersion(), $this->review->getApprovals($userId));
        // if review touches project(s) then
        //  - only members and the review author can change state or attach commits
        //
        // if review touches moderated branch(es) then
        //  - only moderators can approve or reject the review
        //  - authors who are not moderators can move between needs-review/needs-revision/archived and
        //    can attach commits; they cannot approve or reject review
        //  - members can move between needs-review/needs-revision and can attach
        //    commits; they cannot approve/reject or archive
        //  - users that are not project members, moderators or the author cannot perform any transitions
        if (count($moderators)) {
            // For super anything goes from the transitions that have already been established so we
            // don't have to work through the moderator/author allowed states
            if (!$isSuper) {
                $currentState = $this->review->get('state');
                $authorStates = array(Review::STATE_NEEDS_REVIEW, Review::STATE_NEEDS_REVISION, Review::STATE_ARCHIVED);
                $memberStates = array(Review::STATE_NEEDS_REVIEW, Review::STATE_NEEDS_REVISION);
                if ($currentState === Review::STATE_APPROVED) {
                    $authorStates[] = Review::STATE_APPROVED;
                    $authorStates[] = Review::STATE_APPROVED_COMMIT;
                    $memberStates[] = Review::STATE_APPROVED;
                    $memberStates[] = Review::STATE_APPROVED_COMMIT;
                }

                // if the author is also a moderator, treat them as moderator
                if ($isAuthor && !$isModerator) {
                    $this->transitions = in_array($currentState, $authorStates)
                        ? array_intersect_key($this->transitions, array_flip($authorStates))
                        : array();
                } elseif ($isMember && !$isAuthor && !$isModerator) {
                    $this->transitions = in_array($currentState, $memberStates)
                        ? array_intersect_key($this->transitions, array_flip($memberStates))
                        : array();
                } elseif (!$isModerator) {
                    $this->transitions = false;
                }
                $services = $this->options[Option::SERVICES];
                // If this is the world of discrete moderation, then more states need to be filtered
                $moderatorMode =
                    ConfigManager::getValue(
                        $services->get(Option::CONFIG),
                        ConfigManager::REVIEWS_MODERATOR_APPROVAL
                    );
                if (is_array($this->transitions) && ConfigManager::VALUE_EACH === $moderatorMode) {
                    // Moderators from each branch need to approve
                    $canCommit = $this->canCommit($this->review, $moderators);
                    // Allow anything that is not approved, or when approval is allowed
                    $this->transitions = array_flip(
                        array_filter(
                            array_flip($this->transitions),
                            function ($state) use ($isApprover, $canCommit) {
                                $allowed = true;
                                switch ($state) {
                                    case Review::STATE_APPROVED:
                                        $allowed = !$isApprover;
                                        break;

                                    case Review::STATE_APPROVED_COMMIT:
                                        $allowed = $canCommit;
                                        break;
                                    default:
                                        break;
                                }
                                return $allowed;
                            }
                        )
                    );
                }
            }
        } elseif (count($impacted) && !$isMember && !$isAuthor && !$isSuper) {
            $this->transitions = false;
        }
    }

    /**
     * Determine if it is possible to commit the review (assumes that each of the moderators must approve rather
     * than any single moderator having the authority). For commit to be possible
     * Either
     *  - there must be no moderators
     * or
     *  - if there are moderators each of the moderators must have approved the latest revision.
     * The current users approval is presumed when working out if commit is possible. For example if there are no
     * approvals at all but the current users approval would make it possible to commit then the approve/commit
     * option needs to be shown. This is similar to how we handle implicit up votes when checking if a review has
     * votes outstanding.
     * @param Review        $review     the review
     * @param array         $moderators representation of moderators
     * @return bool - true/false => approval states [are/are not] allowed
     */
    private function canCommit(Review $review, $moderators)
    {
        $approvals = $this->getReviewApprovals($review);
        // If there are no approvals, but there are moderators approval
        $canCommit = count($moderators) === 0;
        if (!$canCommit) {
            // There must be approvals from each branch with moderators
            foreach ((array)$moderators as $project => $branches) {
                foreach ($branches as $branch => $branchModerators) {
                    foreach ($approvals as $approver => $versions) {
                        foreach ($branchModerators as $moderator) {
                            if ($moderator === $approver && in_array($review->getHeadVersion(), $versions)) {
                                unset($moderators[$project][$branch]);
                                if (isset($moderators[$project]) && count($moderators[$project]) === 0) {
                                    unset($moderators[$project]);
                                }
                            }
                        }
                    }
                }
            }
            // approved[:committed] is allowed if there are no more moderators needed
            $canCommit = count($moderators) === 0;
        }
        return $canCommit;
    }

    /**
     * Get approvals for the given review and add in an implicit approval from the current
     * user to allow them to see approve and commit if they are the first to ever approve that
     * would mean commit is possible
     * @param Review $review    the current review
     * @return array the approvals
     */
    private function getReviewApprovals(Review $review)
    {
        $approvals = $review->getApprovals();
        // If there are no approvals for the current user or if they have not approved
        // the latest version pretend that they have to work out if approve:commit should
        // be allowed
        if (isset($approvals[$this->options[Option::USER_ID]])) {
            $versions = $approvals[$this->options[Option::USER_ID]];
            if (!in_array($review->getHeadVersion(), $versions)) {
                $versions[]                                 = $review->getHeadVersion();
                $approvals[$this->options[Option::USER_ID]] = $versions;
            }
        } else {
            $approvals[$this->options[Option::USER_ID]] = array($review->getHeadVersion());
        }
        return $approvals;
    }
}
