<?php
/**
 * Functions to help with login.
 *
 * @copyright   2013-2019 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2019.1/1829459
 */

namespace Users\Authentication;

use Application\Config\ConfigException;
use Application\Config\ConfigManager;
use P4\Connection\ConnectionInterface;
use P4\Connection\Exception\ServiceNotFoundException;
use Projects\Model\Project;
use Record\Exception\NotFoundException;
use Users\Model\User;
use Zend\Authentication\Result;
use Zend\Http\Request;
use Zend\Http\Response;
use Zend\View\Model\JsonModel;

class Helper
{
    // Session values
    const COOKIE_LIFETIME            = 'cookie_lifetime';
    const REMEMBERED_COOKIE_LIFETIME = 'remembered_cookie_lifetime';
    const REMEMBER                   = 'remember';
    const SESSION                    = 'session';
    const USERNAME                   = 'username';
    const PASSWORD                   = 'password';
    const ADAPTER                    = 'adapter';
    const ERROR                      = 'error';
    const IS_VALID                   = 'isValid';
    const AUTH_USER                  = 'authUser';
    const SAML_REQUEST               = 'SAMLResponse';
    const RELAY_STATE                = 'RelayState';
    const LOG_ID                     = 'Users:Authentication:Helper:: ';
    const PARAM_REDIRECT             = 'redirect';

    /**
     * Get authenticated user object and invalidate user cache if
     * authenticated user is not in cache - this most likely means
     * that user has been added to Perforce but the form-commit
     * user trigger was not fired
     * @param string $authUser the user identifier
     * @param ConnectionInterface $p4Admin admin connection
     * @return \Users\Model\PluralAbstract
     */
    public static function invalidateCache($authUser, ConnectionInterface $p4Admin)
    {
        // Get authenticated user object and invalidate user cache if
        // authenticated user is not in cache - this most likely means
        // that user has been added to Perforce but the form-commit
        // user trigger was not fired
        if (!User::exists($authUser, $p4Admin)) {
            try {
                $p4Admin->getService('cache')->invalidateItem('users');
            } catch (ServiceNotFoundException $e) {
                // No cache? nothing to invalidate
            }
        }
        return User::fetchById($authUser, $p4Admin);
    }

    /**
     * Attempts to authenticate each candidate in turn breaking at the first
     * success.
     * @param mixed $services
     * @param string $userName the password
     * @param string $password the user name
     * @param bool   $saml whether to use Saml
     * @return array results
     * @throws \Application\Config\ConfigException
     * @see Helper::getLoginCandidates()
     */
    public static function authenticateCandidates($services, $userName, $password, $saml = false)
    {
        $config   = $services->get('config');
        $p4Admin  = $services->get('p4_admin');
        $auth     = $services->get('auth');
        $logger   = $services->get('logger');
        $authUser = null;
        $adapter  = null;
        $error    = null;
        // loop through all login candidates, stop on first success
        foreach (static::getLoginCandidates($services, $userName) as $user) {
            $logger->info('Trying to log in as ' . $user);
            $adapter = new Adapter($user, $password, $p4Admin, $config, $saml);

            try {
                // break if we hit a working candidate
                if ($auth->authenticate($adapter)->getCode() === Result::SUCCESS) {
                    $authUser = $user;
                    $logger->info('Authenticated user ' . $user);
                    break;
                }
            } catch (\Exception $e) {
                // we skip any failed accounts; better luck next try :)
                $logger->info('Attempt to log in as ' . $user . ' failed with ' . $e->getMessage());
                $error = $e->getMessage();
            }
        }
        return array(
            self::AUTH_USER => $authUser,
            self::USERNAME  => $userName,
            self::ADAPTER   => $adapter,
            self::IS_VALID  => $authUser !== null,
            self::ERROR     => $error
        );
    }


    /**
     * Gets a list of login candidates. In most cases it will simply be the user name. If the user name
     * happens to represent an email address (contains '@') then the list returned will contain all the
     * user ids that match that email address. Any user names that are blocked by configuration settings
     * are filtered out.
     * @param mixed     $services
     * @param string    $userName   user name provided to the request
     * @return array    list of candidate users
     * @throws \Application\Config\ConfigException
     */
    public static function getLoginCandidates($services, $userName)
    {
        $config  = $services->get('config');
        $p4Admin = $services->get('p4_admin');
        $blocked = ConfigManager::getValue($config, ConfigManager::SECURITY_PREVENT_LOGIN, array());

        // If we are passed an email (anything with an @) find all matching accounts.
        // Otherwise, simply fetch the passed id.
        $candidates = array();
        if (strpos($userName, '@')) {
            foreach (User::fetchAll(null, $p4Admin) as $candidate) {
                if ($candidate->getEmail() === $userName) {
                    $candidates[] = $candidate->getId();
                }
            }
        } else {
            $candidates[] = $userName;
        }
        return array_diff($candidates, $blocked);
    }

    /**
     * Determines and sets cookie information for a successful login.
     * @param mixed $services
     * @param mixed $request
     * @param array $data containing details relevant to setting cookies
     * @throws \Application\Config\ConfigException
     */
    public static function setCookieInformation($services, $request, $data)
    {
        $path    = $request->getBaseUrl();
        $config  = $services->get('config');
        $session = $services->get(self::SESSION);
        // Remember was an option passed from the 'old' login. Not sure if it will
        // be kept so we check to see if we have it
        $remember = isset($data[self::REMEMBER]) ?: null;
        // The 'remember' setting may have changed; ensure the session cookie
        // is set for the appropriate lifetime before we regenerate its id
        $config[self::SESSION] +=
            array(self::REMEMBERED_COOKIE_LIFETIME => null, self::COOKIE_LIFETIME => null);
        $session->getConfig()->setStorageOption(
            self::COOKIE_LIFETIME,
            $config[self::SESSION][$remember ? self::REMEMBERED_COOKIE_LIFETIME : self::COOKIE_LIFETIME]
        );
        // Regenerate our id since they logged in; this avoids session fixation and also
        // allows any lifetime changes to take affect.
        // As the session was already started there's a Set-Cookie entry for it
        // and regenerating would normally add a second. to avoid two entries (harmless
        // but off-putting) we first clear all Set-Cookie headers.
        header_remove('Set-Cookie');
        session_regenerate_id(true);
        $strict = ConfigManager::getValue($config, ConfigManager::SECURITY_HTTPS_STRICT, false);
        $https  = $request instanceof Request && $request->getUri()->getScheme() == 'https';
        if ($remember) {
            // This cookie sticks around for a year. We don't use the session lifetime
            // here as you want the user id to fill in when the session expires (if remember
            // me was checked). if we shared lifetimes with the session, the user id would
            // never be auto-filled/remembered when you actually needed it.
            $expires = time() + 365 * 24 * 60 * 60;
            headers_sent() ?: setcookie(
                self::REMEMBER,
                $data[self::USERNAME],
                $expires,
                $path,
                '',
                $strict || $https,
                true
            );
        } elseif (isset($_COOKIE[self::REMEMBER])) {
            headers_sent() ?: setcookie(
                self::REMEMBER,
                null,
                -1,
                $path,
                '',
                $strict || $https,
                true
            );
        }
    }

    /**
     * Builds an error message for failed login based on whether it was a user
     * issue or a Swarm configured user issue.
     * @param mixed                 $services
     * @param mixed                 $response
     * @param ConnectionInterface   $p4Admin
     * @param string                $userName
     * @return null
     */
    public static function createErrorMessage($services, $response, ConnectionInterface $p4Admin, $userName)
    {
        $translator = $services->get('translator');
        $logger     = $services->get('logger');
        $error      = null;
        // Login has failed - most likely due to the user supplying bad credentials
        // but there may have been Swarm admin connection changes that if invalid
        // would always stop any user logging in. If we can still use the admin
        // connection then the user details were wrong
        $logger->info($translator->t("Failed login attempt for user '[%s]'", array($userName)));
        try {
            User::fetchById($p4Admin->getUser(), $p4Admin);
            $error = $translator->t("User name or password incorrect.");
        } catch (\Exception $le) {
            $logger->err(
                $translator->t(
                    "An error has occurred with the P4 admin connection '[%s]'",
                    array($le->getMessage())
                )
            );
            $error = $translator->t(
                "There is a configuration problem. Please contact your Swarm administrator."
            );
        }
        $response->setStatusCode(Response::STATUS_CODE_401);
        return $error;
    }

    /**
     * Handles the callback from an IDP to authenticate the Saml response
     * with the P4 connection (and hence the configured trigger)
     * @param mixed $services
     * @return array results
     */
    public static function handleSamlResponse($services)
    {
        $error   = null;
        $userId  = null;
        $adapter = null;
        $logger  = $services->get('logger');
        try {
            $saml = $services->get('php-saml');
            // Need to process the response to get the user name
            $saml->processResponse(null);
            $userId = $saml->getNameId();
            $errors = $saml->getErrors();
            if (empty($errors)) {
                // We still want to authenticate with candidates in case an email was returned
                // by the IDP
                return static::authenticateCandidates($services, $userId, '', true);
            } else {
                $error = $saml->getLastErrorReason();
            }
        } catch (\Exception $e) {
            $error = $e->getMessage();
            $logger->err("An error occurred using Saml login " . $error);
        }
        return array(
            self::AUTH_USER => $userId,
            self::USERNAME  => $userId,
            self::ADAPTER   => $adapter,
            self::ERROR     => $error,
            self::IS_VALID  => $error === null);
    }

    /**
     * Checks the server configuration to see if SSO is enabled.
     * @param mixed $services the services
     * @param mixed $serverId the server id to test. Defaults to null
     * @return bool true if sso_enabled is set to true for the server configuration
     * @throws \Application\Config\ConfigException
     */
    public static function isSSOEnabled($services, $serverId = null)
    {
        $config = $services->get('config');
        $path   = ConfigManager::P4_SSO_ENABLED;
        if ($serverId) {
            $path = ConfigManager::P4 . '.' . $serverId . '.' . ConfigManager::SSO_ENABLED;
        }
        try {
            return ConfigManager::getValue($config, $path, false);
        } catch (ConfigException $e) {
            if ($e->getCode() === ConfigException::PATH_DOES_NOT_EXIST) {
                // This can be expected
                return false;
            }
            throw $e;
        }
    }

    /**
     * Handles SAML login using the PHP SAML library to redirect to the IDP.
     * @param mixed     $services   the services
     * @param mixed     $request    the current request
     * @return JsonModel returns a model with the URL for redirect if a redirect parameter on the request
     * specifies false
     */
    public static function handleSamlLogin($services, $request)
    {
        $logger = $services->get('logger');
        $logger->debug(Helper::LOG_ID . 'getting php-saml library');
        $saml = $services->get('php-saml');
        $logger->debug(Helper::LOG_ID . 'php-saml library success');
        $logger->debug(Helper::LOG_ID . 'Referrer ' . $request->getServer('HTTP_REFERER'));
        // After the IDP has returned we set up to return to the page
        // that we were on originally. For example if we login from
        // the reviews page we want to end up on the reviews page
        $redirect = $request->getQuery(Helper::PARAM_REDIRECT);
        if ($redirect === false || $redirect === 'false') {
            $logger->debug(Helper::LOG_ID . 'saml login without redirect');
            $url = $saml->login($request->getServer('HTTP_REFERER'), array(), false, false, true);
            return new JsonModel(array('url' => $url, 'isValid' => true));
        } else {
            $logger->debug(Helper::LOG_ID . 'login with redirect');
            $saml->login($request->getServer('HTTP_REFERER'));
            new JsonModel(array('isValid' => true));
        }
    }

    /**
     * Gets the referrer that initiated the login request. If that referrer is
     * the actual login api call then null is returned to let the caller decide
     * where to go to avoid a potential looping call between Swarm and an IDP.
     * @return string the value of the referrer if set and not the API call path,
     * otherwise null
     */
    public static function getLoginReferrer()
    {
        $referrer = null;
        if ($_REQUEST && isset($_REQUEST[self::RELAY_STATE])) {
            $loginMatch = preg_match('/.*\/api\/v[0-9]+\/login\/saml*/', $_REQUEST[self::RELAY_STATE]);
            if (!$loginMatch) {
                $referrer = $_REQUEST[self::RELAY_STATE];
            }
        }
        return $referrer;
    }

    /**
     * Utility to handle log out for web and api controllers. Will attempt to redirect to the referrer or
     * a custom logout page unless the parameter 'redirect=false' is requested.
     * @param mixed     $services   application services
     * @param Request   $request    current request
     * @param Response  $response   current response
     * @return Response
     * @throws ConfigException
     */
    public static function logout($services, Request $request, Response $response)
    {
        $logger  = $services->get('logger');
        $session = $services->get('session');
        $auth    = $services->get('auth');

        // clear identity and all other session data on logout
        // note we need to explicitly restart the session (it's closed by default)
        $session->start();
        $auth->clearIdentity();
        $session->destroy(array('send_expire_cookie' => true, 'clear_storage' => true));
        $session->writeClose();
        $redirect = $request->getQuery(Helper::PARAM_REDIRECT);
        if ($redirect === false || $redirect === 'false') {
            $logger->debug(Helper::LOG_ID . 'logged out without redirect');
            return $response;
        } else {
            $logger->debug(Helper::LOG_ID . 'logging out with redirect');
            return Helper::customRedirect($services, $request, $response);
        }
    }

    /**
     * Perform a redirect to the URL
     * @param mixed     $services   application services
     * @param mixed     $url        the url
     * @param Response  $response   the response
     * @return mixed
     */
    private static function redirect($services, $url, Response $response)
    {
        $logger = $services->get('logger');
        $logger->debug(Helper::LOG_ID . 'log out redirect url [' . $url . ']');
        $response->getHeaders()->addHeaderLine('Location', $url);
        $response->setStatusCode(302);
        return $response;
    }

    /**
     * Redirects on logout to either the referrer or a custom logout url if set in config.
     * @param mixed     $services   application services
     * @param Request   $request    current request
     * @param Response  $response   current response
     * @return mixed
     * @throws ConfigException
     */
    private static function customRedirect($services, Request $request, Response $response)
    {
        $logger       = $services->get('logger');
        $config       = $services->get('config');
        $customLogout = ConfigManager::getValue($config, ConfigManager::ENVIRONMENT_LOGOUT_URL);

        if ($customLogout) {
            return static::redirect($services, $customLogout, $response);
        } else {
            // If a referrer is set and it appears to point at us; i want to go to there
            $referrer = $request->getServer('HTTP_REFERER');
            if (strpos($referrer, $request->getUriString()) !== false) {
                // Redirecting to here would result in a circular redirect error
                $logger->warn(Helper::LOG_ID . "Averting circular redirect to  $referrer");
                return $response;
            };
            $host = $request->getUri()->getScheme() . '://' . $request->getUri()->getHost();
            // Check the url address for project in it if found check project really exist and then check if its private
            preg_match('#projects\/(.*)#', $referrer, $matches);
            $projectID = end($matches);
            if ($projectID) {
                $p4Admin = $services->get('p4_admin');
                try {
                    // If the project is private we do not want to redirect to the project on logout
                    $referrer = (Project::fetch($projectID, $p4Admin)->isPrivate()) ? '/' : $referrer;
                } catch (NotFoundException $e) {
                } catch (\InvalidArgumentException $e) {
                }
            }
            $logger->debug(Helper::LOG_ID . 'redirect referrer [' . $referrer . ']');
            if ($referrer !== '/' && stripos($referrer, $host) === 0) {
                return static::redirect($services, $referrer, $response);
            }
        }
        return static::redirect($services, '/', $response);
    }
}
